ALPHABET = "abcdefghijklmnopqrstuvwxyz"
DISTRIBUTIONS = [
    "exponential",
    "gumbel",
    "laplace",
    "logistic",
    "lognormal",
    "normal",
    "poisson",
    "rayleigh",
    "standard_cauchy",
    "standard_exponential",
    "standard_normal",
    "uniform"
]
